/*
!C******************************************************************
!Description:

	Functions related to processing of each granule.

!Revision History:
!Team-unique Header:
!References and Credits:

	L. Giglio, Jan. 2001

!END*************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>


#include "mfhdf.h"


#include "const.h"
#include "types.h"
#include "tb.h"
#include "fire.h"
#include "detfire.h"
#include "firetab.h"
#include "io.h"
#include "process.h"
#include "cmg.h"
#include "zstats.h"
#include "error.h"


static void clear_scan(Scan *scan);
static void update_cmg(OutScan *outscan, Scan *scan);



/* ---------- global flags ---------- */

extern int verbose;
extern int write_coarse;
extern int write_cmg;



int ProcGranule(HDF_InputInfo *hinfo, HDF_OutputInfo *hfire,
	FireCounts *counts, int skip_water)
{
	int i;
	int iscan;
	int st;
	Scan *PrevPtr, *CurrPtr, *NextPtr;
	Scan *save;

	Scan *scanptr[3];	/*  0 = previous, 1 = current, 2 = next */

	OutScan *outscan;

	int first_line;

	ScanThreshPF *pfthresh;


	/* allocate structure to hold potential fire thresholds for scan */
	if ( !(pfthresh = (ScanThreshPF *) malloc(sizeof(ScanThreshPF))) )
		return error(ERR_F_MEM);

	/* allocate buffer for output */
	if ( !(outscan = (OutScan *) malloc(sizeof(OutScan))) )
		return error(ERR_F_MEM);

	/* allocate triple buffer */
	for (i = 0; i < 3; i++) {
		if ( !(scanptr[i] = (Scan *) malloc(sizeof(Scan))) )
			return error(ERR_F_MEM);
		}

	PrevPtr = scanptr[0];
	CurrPtr = scanptr[1];
	NextPtr = scanptr[2];

	init_zstats();

	/* zero pixel counts */
	(void) memset(counts, 0, sizeof(FireCounts));

	/* handle zero scans */
	if (hinfo->nscans < 1) {
		if ( (st = write_fire_attributes(hfire->sd_id_fire, counts)) )
			return st;

		if (write_coarse)
			if ( (st = write_fire_attributes(hfire->sd_id_coarse,
				counts)) )
				return st;

		/* free triple buffer */
		for (i = 0; i < 3; i++) free(scanptr[i]);

		free_zstats();

		return 0;
		}

	new_fire_pixel_table();


	/* clear previous buffer */
	clear_scan(PrevPtr);

	/* fill current buffer */
	st = ReadScan(hinfo, 0, CurrPtr);
	if (st) return st;


	/* screen current buffer */
	prescreen(CurrPtr, skip_water);


	first_line = 0;
	for (iscan = 1; iscan < hinfo->nscans; iscan++) {
#ifdef DEBUG
		if (verbose) if (!(iscan % 10)) printf("%d\n", iscan);
#endif

		/* fill next buffer */
		st = ReadScan(hinfo, iscan, NextPtr);
		if (st) return st;


		/* screen next buffer */
		prescreen(NextPtr, skip_water);


#ifdef DEBUG
printf("First line %d\n", first_line);
#endif
		/* ----- detect fires in current scan (CurrPtr) ----- */
		compute_zone_stats(PrevPtr, CurrPtr, NextPtr);

		/* pre-compute potential fire thresholds for each sample */
		compute_scan_PF_thresholds(CurrPtr, pfthresh);

		/* flag potential fire pixels in current scan (the contextual
		detection will only be performed on this buffer) */
		flag_potential_fires(CurrPtr, pfthresh);

		/* unlike potential fire pixels, we must flag
		background fire pixels in all three buffers because the
		contextual detection phase will use observations from all of
		them */
		flag_background_fires(PrevPtr);
		flag_background_fires(CurrPtr);
		flag_background_fires(NextPtr);

		st = detect(PrevPtr, CurrPtr, NextPtr, first_line, skip_water,
			pfthresh, outscan, counts);
		if (st) return st;

		if ( (st = WriteScan(hfire, iscan - 1, outscan, CurrPtr,
			pfthresh)) ) return st;

		if (write_cmg) update_cmg(outscan, CurrPtr);


		save = PrevPtr;
		PrevPtr = CurrPtr;
		CurrPtr = NextPtr;
		NextPtr = save;

		/*
		At this point scan buffer pointed to by NextPtr contains data
		from an earlier scan that will be overwritten.
		*/

		first_line += LINES_PER_SCAN_1KM;
		}

	clear_scan(NextPtr);

	/* do detection on current buffer */
	compute_zone_stats(PrevPtr, CurrPtr, NextPtr);

	compute_scan_PF_thresholds(CurrPtr, pfthresh);

	flag_potential_fires(CurrPtr, pfthresh);

	flag_background_fires(PrevPtr);
	flag_background_fires(CurrPtr);

	st = detect(PrevPtr, CurrPtr, NextPtr, first_line, skip_water,
		pfthresh, outscan, counts);
	if (st) return st;

	if ( (st = WriteScan(hfire, iscan - 1, outscan, CurrPtr,
		pfthresh)) ) return st;
	if (write_cmg) update_cmg(outscan, CurrPtr);


	free(pfthresh);
	free(outscan);

	/* free triple buffer */
	for (i = 0; i < 3; i++) free(scanptr[i]);

	free_zstats();

	st = write_fire_pixel_table(hfire->sd_id_fire, write_cmg);
	if (st) return st;

	counts->NumPix = hinfo->nscans * LINES_PER_SCAN_1KM * SAMPLES_1KM;

	st = write_fire_attributes(hfire->sd_id_fire, counts);
	if (st) return st;

	if (write_coarse)
		if ( (st = write_fire_attributes(hfire->sd_id_coarse, counts)) )
			return st;

	return 0;
}



void compute_reflectances(float R1[LINES_PER_SCAN_1KM][SAMPLES_1KM],
	float R2[LINES_PER_SCAN_1KM][SAMPLES_1KM],
	float R7[LINES_PER_SCAN_1KM][SAMPLES_1KM],
	float SolZenAng[LINES_PER_SCAN_1KM][SAMPLES_1KM])
{
	int i, j;
	float mu_s;
	float inv_mu_s;


	for (i = 0; i < LINES_PER_SCAN_1KM; i++) {
		for (j = 0; j < SAMPLES_1KM; j++) {
			/* handle nighttime observations */
			if (SolZenAng[i][j] > 85.0F) {
				R1[i][j] = R2[i][j] = -1.0F;
				continue;
				}

			mu_s = cosf(D2R * SolZenAng[i][j]);
			inv_mu_s = 1.0F / mu_s;

			if (R1[i][j] >= 0.0F)
				R1[i][j] *= inv_mu_s;

			if (R2[i][j] >= 0.0F)
				R2[i][j] *= inv_mu_s;

			if (R7[i][j] >= 0.0F)
				R7[i][j] *= inv_mu_s;
			}
		}
}




void compute_tb(float L[LINES_PER_SCAN_1KM][SAMPLES_1KM],
	float Tb[LINES_PER_SCAN_1KM][SAMPLES_1KM], float lambda)
/*
!C*****************************************************************************
!Description:
	Compute brightness temperatures for buffered radiances.

!Input Parameters:
	L	radiance (W m^-2 sr^-1 micrometer^-1)
	lambda	central wavelength (micrometers)

!Output Parameters:
	Tb	brightness temperature (K)

!Input/Output Parameters:
	None.

!Returns:
	Nothing.

!Revision History
	Original version February 2001, L. Giglio.

!Team-unique Header:

!References and Credits

!Design Notes

!END***************************************************************************
*/
{
	int i, j;
	float lambda5;

	/* precompute lambda^5 for speed */
	lambda5 = powf(lambda, 5.0F);


	for (i = 0; i < LINES_PER_SCAN_1KM; i++) {
		for (j = 0; j < SAMPLES_1KM; j++) {
			if (L[i][j] > 0.0F)
				Tb[i][j] = FastInversePlanck(L[i][j], lambda,
					lambda5);
			else
				Tb[i][j] = -1.0F;
			}
		}
}




void correct_tb(float Tb[LINES_PER_SCAN_1KM][SAMPLES_1KM], float *c)
/*
!C*****************************************************************************
!Description:
	Correct brightness temperatures computing via central wavelength
	for spectral response.

!Input Parameters:
	c	Correction coefficients; c[0] is slope, c[1] is intercept.

!Output Parameters:
	None.

!Input/Output Parameters:
	Tb	brightness temperature (K)

!Returns:
	Nothing.

!Revision History
	L. Giglio.

!Team-unique Header:

!References and Credits

!Design Notes

!END***************************************************************************
*/
{
	int i, j;

	for (i = 0; i < LINES_PER_SCAN_1KM; i++) {
		for (j = 0; j < SAMPLES_1KM; j++) {
			if (Tb[i][j] > 0.0F)
				Tb[i][j] = Tb[i][j] * c[0] + c[1];
			}
		}
}



/* fill all fields of specified scan with missing data values */
static void clear_scan(Scan *scan)
/*
!C*****************************************************************************
!Description:
	Fill specified scan buffer with missing data values.

!Input Parameters:
	None.

!Output Parameters:
	scan	pointer to scan

!Input/Output Parameters:
	None.

!Returns:
	Nothing.

!Revision History
	Original version January 2001, L. Giglio.

!Team-unique Header:

!References and Credits

!Design Notes

!END***************************************************************************
*/
{
	int i, j;

	/* missing data value used for radiance, reflectance, and brightness
	temperature fields */
	const float missval = -1.0F;

	PixInfo NullPixInfo;


	(void) memset(&NullPixInfo, 0, sizeof(PixInfo));

	for (i = 0; i < LINES_PER_SCAN_1KM; i++) {
		for (j = 0; j < SAMPLES_1KM; j++) {
			scan->L21[i][j] = scan->L22[i][j] = missval;
			scan->T21[i][j] = scan->T22[i][j] = missval;
			scan->T31[i][j] = scan->T32[i][j] = missval;
			scan->R1[i][j] = scan->R2[i][j] = scan->R7[i][j] =
				missval;

			scan->lat[i][j] = scan->lon[i][j] = -999.9F;

			scan->SolZenAng[i][j] = scan->ViewZenAng[i][j] =
				scan->SolAzAng[i][j] = scan->ViewAzAng[i][j] =
				-999.0F;

			scan->landsea[i][j] = 0;

			scan->pixflags[i][j] = NullPixInfo;
			}
		}
}




static void update_cmg(OutScan *outscan, Scan *scan)
{
	int i, j;

	for (i = 0; i < LINES_PER_SCAN_1KM; i++) {
		for (j = 0; j < SAMPLES_1KM; j++)
			cmg_append(scan->lat[i][j], scan->lon[i][j],
				outscan->ExtFireMask[i][j],
				scan->pixflags[i][j]);
		}
}

